/* Maurice.Clerc@WriteMe.com  (mars 1999)
Small arithmetic toolbox for fuzzy numbers

You can easily add your own function (see in particular how FUZZY_SIN is defined)
*/

#include <math.h>
#include <stdio.h>

FILE *trace;
struct fuzzy_number {float f[4][2];int flag;}; /* double trapezium 
					{ (f[0,0],0),(f[1,0],1),(f[2,0],1),(f[0,3],0),(f[1,0],0),(f[1,1],1),(f[2,1],1),(f[3,1],0) }
					f := "positive" part         and g:= "negative" part
					0<=f[0,0]<=f[1,0]<=f[2,0]<=f[3,0]    and      f[0,1]<=f[1,1]<=f[2,1]<=f[3,1]<=0	 
					flag = 0 <=> number 0
					     = 1 <=> just a positive part
						 = 2 <=> just a negative part
						 = 3 <=> both parts                         
				Examples:	the crisp positive number 2 is given as 0 0 0 0 2 2 2 2 
							the crisp negative number -3 is given  as   -3 -3 -3 -3 0 0 0 0
							a "positive" fuzzy number  0 0 0 0 2 2.5 4.5 5
							a fuzzy number "around" 0:  -3 -2 0 0 0 0 4 5				
						
						*/						    
/* Fuzzy functions in alphabetic order */					
struct fuzzy_number	FUZZY_COEFF(struct fuzzy_number a,float coeff);
float				FUZZY_COMPAT(struct fuzzy_number a,struct fuzzy_number b);
struct fuzzy_number	FUZZY_DIFF(struct fuzzy_number a,struct fuzzy_number b);
struct fuzzy_number	FUZZY_DIV(struct fuzzy_number a,struct fuzzy_number b);
struct fuzzy_number	FUZZY_EXP(struct fuzzy_number a);
float				FUZZY_EVAL(struct fuzzy_number a);
struct fuzzy_number	FUZZY_EXP(struct fuzzy_number a);
int					FUZZY_FLAG(struct fuzzy_number a);
void 				FUZZY_FPRINTF(FILE *file,struct fuzzy_number fz);
void 				FUZZY_FSCANF(FILE *file,struct fuzzy_number *);
struct fuzzy_number	FUZZY_INV(struct fuzzy_number a);
struct fuzzy_number	FUZZY_LOG(struct fuzzy_number a);
struct fuzzy_number	FUZZY_MAX(struct fuzzy_number a,struct fuzzy_number b);
struct fuzzy_number	FUZZY_MIN(struct fuzzy_number a,struct fuzzy_number b);

struct fuzzy_number	FUZZY_OPP(struct fuzzy_number a);
struct fuzzy_number	FUZZY_POW(struct fuzzy_number a,float p);
void				FUZZY_PRINTF(struct fuzzy_number fz);
struct fuzzy_number	FUZZY_PROD(struct fuzzy_number a,struct fuzzy_number b);
void				FUZZY_SCANF(struct fuzzy_number *);
struct fuzzy_number	FUZZY_SIN(struct fuzzy_number a);
struct fuzzy_number	FUZZY_SORT(struct fuzzy_number a);
struct fuzzy_number	FUZZY_SUM(struct fuzzy_number a,struct fuzzy_number b);

/* Other functions */
float				MIN(float a,float b);
float				MAX(float a,float b);

/* Other declarations */
float				infinity=10000;
float				zero=0.0001;
int					eval_type=3; /* See subroutine FUZZY_EVAL */
struct fuzzy_number	fuzzy_infinity={infinity,0,infinity,0,infinity,0,infinity,0,1};
struct fuzzy_number	fuzzy_NaN={sqrt(-1),sqrt(-1),sqrt(-1),sqrt(-1),sqrt(-1),sqrt(-1),sqrt(-1),sqrt(-1),3};
struct fuzzy_number	fuzzy_null={0,0,0,0,0,0,0,0,0};
struct fuzzy_number	fuzzy_one={1,0,1,0,1,0,1,0,1};




main()  /* Just testing the toolbox */
{
//FILE *data;
//FILE *trace;

struct fuzzy_number a;
struct fuzzy_number b;
struct fuzzy_number fz;

char				again;
float				coeff,f,p;


//data=fopen("DATA.TXT","r");
//trace=fopen("TRACE.TXT","w");


input_loop:


printf("A (8 values) ?: ");FUZZY_SCANF(&a);
printf("\nB (8 values) ?: ");FUZZY_SCANF(&b);
printf("power p ? ");scanf("%f",&p);
printf("coeff ? ");scanf("%f",&coeff);
printf("\n");


/*   if data are read from a file
FUZZY_FSCANF(data,&a);
FUZZY_FSCANF(data,&b);
fscanf("%f",&p);
fscanf("%f",&coeff);
*/


/* an example of how to "print" results on a file
fprintf(trace,"a: ");FUZZY_FPRINTF(trace,a);
fprintf(trace,"\neval a: ");fprintf(trace,"%f",FUZZY_EVAL(a));
printf ("\na: ");FUZZY_PRINTF(a);
printf("\n");
*/

fz=FUZZY_COEFF(a,coeff);
printf ("\ncoeff*a: ");FUZZY_PRINTF(fz);printf("\n");

f=FUZZY_COMPAT(a,b);
printf("\ncompatibility (a,b): %f ",f);printf("\n");

fz=FUZZY_DIFF(a,b);
printf ("\na-b: ");FUZZY_PRINTF(fz);
printf ("  eval(a)-eval(b) %f ",FUZZY_EVAL(a)-FUZZY_EVAL(b));
printf("\n");

fz=FUZZY_DIV(a,b);
printf ("\na/b: ");FUZZY_PRINTF(fz);
printf ("  eval(a)/eval(b) %f ",FUZZY_EVAL(a)/FUZZY_EVAL(b));
printf("\n");

fz=FUZZY_EXP(a);
printf ("\n(exp(a): ");FUZZY_PRINTF(fz);printf("\n");


fz=FUZZY_INV(a);
printf ("\n1/a: ");FUZZY_PRINTF(fz);
printf ("  eval(a)*eval(1/a) %f ",FUZZY_EVAL(a)*FUZZY_EVAL(fz));
printf("\n");


fz=FUZZY_LOG(a);
printf ("\n(log(a): ");FUZZY_PRINTF(fz);printf("\n");

fz=FUZZY_MAX(a,b);
printf ("\nmax(a,b): ");FUZZY_PRINTF(fz);printf("\n");

fz=FUZZY_MIN(a,b);
printf ("\nmin(a,b): ");FUZZY_PRINTF(fz);printf("\n");

fz=FUZZY_OPP(a);
printf ("\n-a: ");FUZZY_PRINTF(fz);
printf ("  eval(a)+eval(-a) %f ",FUZZY_EVAL(a)+FUZZY_EVAL(fz));
printf("\n");

fz=FUZZY_POW(a,p);
printf ("\n(a^p: ");FUZZY_PRINTF(fz);printf("\n");

fz=FUZZY_PROD(a,b);
printf ("\na*b: ");FUZZY_PRINTF(fz);
printf ("  eval(a)*eval(b) %f ",FUZZY_EVAL(a)*FUZZY_EVAL(b));
printf("\n");

fz=FUZZY_SIN(a);
printf ("\nsin(a): ");FUZZY_PRINTF(fz);printf("\n");

fz=FUZZY_SUM(a,b);
printf ("\na+b: ");FUZZY_PRINTF(fz);
printf ("  eval(a)+eval(b) %f ",FUZZY_EVAL(a)+FUZZY_EVAL(b));
printf("\n");



printf("\nAgain ?(Y/y): ");
scanf("%s",&again);
if (again=='Y' || again=='y') goto input_loop;

//fclose(data);
//fclose(trace);
}
/*========================= HERE IS THE PACKAGE (alphabetic order) ===========================*/
/*------------------------------------------------------------------------- */
struct fuzzy_number	FUZZY_COEFF(struct fuzzy_number a,float coeff)
/* Multiply "a" by "coeff" */
{struct fuzzy_number ff;int i,j;
if (coeff>=0)
	{
	for (j=0;j<2;j++)
	for (i=0;i<4;i++) ff.f[i][j]=coeff*a.f[i][j];
	}
	else
	{
	for (j=0;j<2;j++)
	for (i=0;i<4;i++) ff.f[3-i][1-j]=coeff*a.f[i][j];
	}
ff.flag=FUZZY_FLAG(ff);
return ff;

}
/*------------------------------------------------------------------------- */
float FUZZY_COMPAT(struct fuzzy_number a,struct fuzzy_number b)
/* Compatibility level in [0,1] */
 {float hp,hn;

/* Strong compatibility if ( core(a) inter core(b) ) non-empty */

if (a.flag==0 &&b.flag==0) return 1; /* special case a=b=0 */
	
/* Positive parts */
if ((a.flag==1 || a.flag==3) && (b.flag==1 || b.flag==3))
	{
	if ((a.f[2][0]-b.f[1][0])*(a.f[2][0]-b.f[2][0])<=0) return 1;
	if ((b.f[2][0]-a.f[1][0])*(b.f[2][0]-a.f[2][0])<=0) return 1;
	}
	/* Negative parts */
if 	((a.flag==2 || a.flag==3) && (b.flag==2 || b.flag==3))
	{
	if ((a.f[2][1]-b.f[1][1])*(a.f[2][1]-b.f[2][1])<=0) return 1;
	if ((b.f[2][1]-a.f[1][1])*(b.f[2][1]-a.f[2][1])<=0) return 1;
	}
/*--------------------- */
/* Weak compatibility if (a inter b) non-empty */
hp=0;hn=0;
/* Positive parts */
if ((a.flag==1 || a.flag==3) && (b.flag==1 || b.flag==3))
	{
	if ((a.f[0][0]-b.f[0][0])*(a.f[0][0]-b.f[3][0])<=0)
	{hp=(a.f[0][0]-b.f[3][0])/((b.f[2][0]-b.f[3][0])-(a.f[1][0]-a.f[0][0]));
	 }
	
	if ((b.f[0][0]-a.f[0][0])*(b.f[0][0]-a.f[3][0])<=0)
	{hp=(b.f[0][0]-a.f[3][0])/((a.f[2][0]-a.f[3][0])-(b.f[1][0]-b.f[0][0]));
	}
	}
	/* Negative parts */
if 	((a.flag==2 || a.flag==3) && (b.flag==2 || b.flag==3))
	{
	if ((a.f[0][1]-b.f[0][1])*(a.f[0][1]-b.f[3][1])<=0)
	{hn=(a.f[0][1]-b.f[3][1])/((b.f[2][1]-b.f[3][1])-(a.f[1][1]-a.f[0][1]));
	}
	
	if ((b.f[0][1]-a.f[0][1])*(b.f[0][1]-a.f[3][1])<=0)
	{hn=(b.f[0][1]-a.f[3][1])/((a.f[2][1]-a.f[3][1])-(b.f[1][1]-b.f[0][1]));
	}
	}
/*--------------------- */
return MAX(hp,hn);
}
/*------------------------------------------------------------------------- */
struct fuzzy_number	FUZZY_DIFF(struct fuzzy_number a,struct fuzzy_number b)
{ struct fuzzy_number ff;
ff=FUZZY_SUM(a,FUZZY_OPP(b));
return ff;
}
/*------------------------------------------------------------------------- */
struct fuzzy_number	FUZZY_DIV(struct fuzzy_number a,struct fuzzy_number b)
{struct fuzzy_number ff;

ff=FUZZY_PROD(a,FUZZY_INV(b));
ff.flag=FUZZY_FLAG(ff);
return ff;
}

/*------------------------------------------------------------------------- */
float	FUZZY_EVAL(struct fuzzy_number a) /* Defuzzification of a double Left-Right fuzzy set 
                                             (C) Maurice Clerc ;-)   */
{
float eval,peval,neval,e1,e2,p1,p2,p3,n1,n2,n3,mp1,mp2,mp3,mn1,mn2,mn3,m;
int evaltype;

if (eval_type>=0 && eval_type<=3) evaltype=eval_type; else evaltype=3;

switch (eval_type)
{
case 0:
	{
	peval=((a.f[0][0]+a.f[3][0])+2*(a.f[1][0]+a.f[2][0]))/6;
	neval=((a.f[0][1]+a.f[3][1])+2*(a.f[1][1]+a.f[2][1]))/6;
	eval=peval+neval;
	break;
	}
case 1:
	{
	peval=(2*sqrt(a.f[1][0]*a.f[2][0])+sqrt(a.f[0][0]*a.f[3][0]))/3;
	neval=-(2*sqrt(a.f[1][1]*a.f[2][1])+sqrt(a.f[0][1]*a.f[3][1]))/3;
	eval=peval+neval;
	break;}
case 2:
	{
	e1=(a.f[1][0]+a.f[2][0])/sqrt(4+pow((a.f[1][0]+a.f[2][0]),2)-4*(a.f[1][0]*a.f[2][0]));
	e2=(a.f[0][0]+a.f[3][0])/sqrt(4+pow((a.f[0][0]+a.f[3][0]),2)-4*(a.f[0][0]*a.f[3][0]));
	peval=(2*e2+e1)/3;

	e1=(a.f[1][1]+a.f[2][1])/sqrt(4+pow((a.f[1][1]+a.f[2][1]),2)-4*(a.f[1][1]*a.f[2][1]));
	e2=(a.f[0][1]+a.f[3][1])/sqrt(4+pow((a.f[0][1]+a.f[3][1]),2)-4*(a.f[0][1]*a.f[3][1]));
	neval=(2*e2+e1)/3;
	eval=peval+neval;
	break;}
case 3: /* gravity center */
	p1=a.f[0][0]+(a.f[1][0]-a.f[0][0])/sqrt(2);mp1=(a.f[1][0]-a.f[0][0])/2;
	p2=(a.f[2][0]+a.f[1][0])/2;mp2=(a.f[2][0]-a.f[1][0]);
	p3=a.f[3][0]-(a.f[3][0]-a.f[2][0])/sqrt(2);mp3=(a.f[3][0]-a.f[2][0])/2;
	
	n1=a.f[0][1]+(a.f[1][1]-a.f[0][1])/sqrt(2);mn1=(a.f[1][1]-a.f[0][1])/2;
	n2=(a.f[2][1]+a.f[1][1])/2;mn2=(a.f[2][1]-a.f[1][1]);
	n3=a.f[3][1]-(a.f[3][1]-a.f[2][1])/sqrt(2);mn3=(a.f[3][1]-a.f[2][1])/2;
	m=mp1+mp2+mp3+mn1+mn2+mn3;
	if(fabs(m)>zero)
		eval=(mp1*p1+mp2*p2+mp3*p3+mn1*n1+mn2*n2+mn3*n3)/m;
	else /* it is a crisp number either positive or negative */
		eval=a.f[0][0]+a.f[0][1];
	
	break;
}	
return eval;
}
/*------------------------------------------------------------------------- */
struct fuzzy_number	FUZZY_EXP(struct fuzzy_number a)
{struct fuzzy_number ff, dff;float n;int i,exp_type; float f[4]={infinity,infinity,0,0};
exp_type=0;
switch (exp_type)
	{
	case 0: /* we use "interval" approach */
	ff=fuzzy_null;
	ff.f[0][0]=infinity;ff.f[1][0]=infinity;

	if (a.flag==1 || a.flag==3) /* "a" has a positive part */
	for (i=0;i<4;i++) ff.f[i][0]=exp(a.f[i][0]);

	if (a.flag>1) /* "a" has a negative part */
	for (i=0;i<4;i++) f[i]=exp(a.f[i][1]);

	ff.f[0][0]=MIN(ff.f[0][0],f[0]);
	ff.f[1][0]=MIN(ff.f[1][0],f[1]);
	ff.f[2][0]=MAX(ff.f[2][0],f[2]);
	ff.f[3][0]=MAX(ff.f[3][0],f[3]);
	break;

	case 1: /* we use a definition of exp as limit of polynomial sum */
	ff=fuzzy_one;dff=a;n=1;
	while (fabs(FUZZY_EVAL(dff))>zero)
		{
		ff=FUZZY_SUM(ff,dff);
		dff=FUZZY_PROD(dff,a);
		n=n+1;
		dff=FUZZY_COEFF(dff,1/n);
	
		}

		/*for (i=0;i<4;i++) {ff.f[i][0]=ff.f[i][0]+ff.f[i][1];ff.f[i][1]=0;}*/
		
}
ff.flag=1;	
return ff;
}
/*------------------------------------------------------------------------- */
int FUZZY_FLAG(struct fuzzy_number a)
{int flagp,flagn;
flagp=0;flagn=0;
if (a.f[3][0]>0) flagp=1;
if (a.f[0][1]<0) flagn=2;
return flagp+flagn;

}
/*------------------------------------------------------------------------- */
void FUZZY_FPRINTF(FILE *file,struct fuzzy_number fz)
{int null=0;

if (fz.flag==0) {fprintf(file,"%i",null);return;}; /* Null number */

if (fz.flag>1) /* there is a negative part */
fprintf(file,"%f %f %f %f",fz.f[0][1],fz.f[1][1],fz.f[2][1],fz.f[3][1]);

if (fz.flag==1 || fz.flag==3) /* there is a positive part */
fprintf(file,"%f %f %f %f ",fz.f[0][0],fz.f[1][0],fz.f[2][0],fz.f[3][0]);


}
/*------------------------------------------------------------------------- */
void FUZZY_FSCANF(FILE *file,struct fuzzy_number * fz)
{
int i,j;struct fuzzy_number ff;

fscanf(file,"%f %f %f %f %f %f %f %f",&ff.f[0][1],&ff.f[1][1],&ff.f[2][1],&ff.f[3][1],&ff.f[0][0],&ff.f[1][0],&ff.f[2][0],&ff.f[3][0]);
ff.flag=FUZZY_FLAG(ff);

for (j=0;j<2;j++)
for (i=0;i<4;i++) {fz->f[i][j]=ff.f[i][j];};

fz->flag=ff.flag;
}
/*------------------------------------------------------------------------- */
struct fuzzy_number	FUZZY_INV(struct fuzzy_number a)
{
int i; struct fuzzy_number ff;
ff=fuzzy_null;
if (a.flag==0) /* This is the null number */
	{
	{
	for (i=0;i<4;i++) ff.f[i][0]=infinity;};
	ff.flag=1;return ff;
	};
	
if (a.flag==1 || a.flag==3) /* There is a positive part */
	{for (i=0;i<4;i++)
		if (fabs(a.f[3-i][0])>zero) {ff.f[i][0]=1/a.f[3-i][0];}
			else ff.f[i][0]=infinity;
	};
		
if (a.flag>1) /* There is a negative part */
	{for (i=0;i<4;i++)
		if (fabs(a.f[3-i][1])>zero) {ff.f[i][1]=1/a.f[3-i][1];}
			else ff.f[i][1]=-infinity;
	};
				
ff.flag=FUZZY_FLAG(ff);		
return ff;
}		

/*------------------------------------------------------------------------- */
struct fuzzy_number	FUZZY_LOG(struct fuzzy_number a)
{struct fuzzy_number ff; int i;

	/* we use "interval" approach 
	We take into account only the positive part (if any) */
	
	if (a.flag==0) return FUZZY_OPP(fuzzy_infinity);
	if (a.flag>1) return fuzzy_NaN; /* Defined only for "positive" fuzzy numbers */
	
	ff=fuzzy_null;
	for (i=0;i<4;i++) 
		if (a.f[i][0]<zero) ff.f[i][1]=-infinity; 
		else
			if (a.f[i][0]<1) ff.f[i][1]=log(a.f[i][0]);
			else ff.f[i][0]=log(a.f[i][0]);
	
	ff.flag=FUZZY_FLAG(ff);
	return ff;
}	
	
/*------------------------------------------------------------------------- */
struct fuzzy_number	FUZZY_MAX(struct fuzzy_number a,struct fuzzy_number b)
{
if (FUZZY_EVAL(a)>FUZZY_EVAL(b)) return a;return b;
}
/*------------------------------------------------------------------------- */
struct fuzzy_number	FUZZY_MIN(struct fuzzy_number a,struct fuzzy_number b)
{
if (FUZZY_EVAL(a)<FUZZY_EVAL(b)) return a; return b;
}
/*------------------------------------------------------------------------- */
struct fuzzy_number	FUZZY_OPP(struct fuzzy_number a)
{
int i,j; struct fuzzy_number ff;

for (j=0;j<2;j++)
for (i=0;i<4;i++) {ff.f[i][1-j]=-a.f[3-i][j];};
ff.flag=FUZZY_FLAG(ff);

return ff;

}
/*------------------------------------------------------------------------- */
struct fuzzy_number	FUZZY_POW(struct fuzzy_number a,float p)
{int i;struct fuzzy_number fz;

if (floor(p)==p) /* If p is an integer, there is no restriction for a */
	{fz=fuzzy_one;
	for (i=0;i< fabs(p);i++) fz=FUZZY_PROD(fz,a);
	if (p<0) fz=FUZZY_DIV(fuzzy_one,fz);
	}
else
	if (a.flag==1) /* Defined only for "positive" fuzzy numbers */
		{
		fz=FUZZY_LOG(a);
		fz=FUZZY_COEFF(fz,p);
		fz=FUZZY_EXP(fz);
		}
	else
		fz=fuzzy_NaN;
return fz;		
}
/*------------------------------------------------------------------------- */
void FUZZY_PRINTF(struct fuzzy_number fz)
{int null=0;float flag;int type;
type=1; /* You  can choose the type you prefer */
switch (type)
{
case 1: /* extended print */
{
flag=FUZZY_FLAG(fz);
printf("%f %f %f %f",fz.f[0][1],fz.f[1][1],fz.f[2][1],fz.f[3][1]);
printf(" %f %f %f %f ",fz.f[0][0],fz.f[1][0],fz.f[2][0],fz.f[3][0]);
printf("\neval: ");printf("%f",FUZZY_EVAL(fz));
return;
}
case 2: /* shorter print */
{
if (fz.flag==0) {printf("%i",null);return;}; /* Null number */


if (fz.flag>1) /* there is a negative part */
printf("%f %f %f %f",fz.f[0][1],fz.f[1][1],fz.f[2][1],fz.f[3][1]);


if (fz.flag==1 || fz.flag==3) /* there is a positive part */
printf("%f %f %f %f ",fz.f[0][0],fz.f[1][0],fz.f[2][0],fz.f[3][0]);

printf("\neval: ");printf("%f",FUZZY_EVAL(fz));
}
}
}
/*------------------------------------------------------------------------- */
struct fuzzy_number	FUZZY_PROD(struct fuzzy_number a,struct fuzzy_number b)
{
struct fuzzy_number ff; int i;


ff=fuzzy_null;

if (a.flag==0 || b.flag==0) return ff;

if ((a.flag==1 && b.flag==1) || ( a.flag==1 && b.flag==3 ) || (a.flag==3 && b.flag==1))
{	/* Positive part */

	for (i=0;i<4;i++) ff.f[i][0]=a.f[i][0]*b.f[i][0];
}

if ((a.flag==2 && b.flag==2) || (a.flag==2 && b.flag==3) || (a.flag==3 && b.flag==2))
{	/* Positive part */

	for (i=0;i<4;i++) ff.f[3-i][0]=a.f[i][1]*b.f[i][1];
}


if ((a.flag==2 && (b.flag==1 || b.flag==3)) || (a.flag==3 && b.flag==1)  )
{	
	/* Negative part */
	
	for (i=0;i<4;i++) ff.f[3-i][1]=a.f[3-i][1]*b.f[i][0];
}

if ((b.flag==2 && (a.flag==1 || a.flag==3)) || (a.flag==1 && b.flag==3)    )
{	
	/* Negative part */
	for (i=0;i<4;i++) ff.f[3-i][1]=b.f[3-i][1]*a.f[i][0];
}


if (a.flag==3 && b.flag==3) /* General case */
{
/* Positive part */
ff.f[0][0]=MIN(a.f[0][0]*b.f[0][0],a.f[3][1]*b.f[3][1]);
ff.f[1][0]=MIN(a.f[1][0]*b.f[1][0],a.f[2][1]*b.f[2][1]);
ff.f[2][0]=MAX(a.f[2][0]*b.f[2][0],a.f[1][1]*b.f[1][1]);
ff.f[3][0]=MAX(a.f[3][0]*b.f[3][0],a.f[0][1]*b.f[0][1]);

/* Negative part */

ff.f[0][1]=MIN(a.f[0][1]*b.f[3][0],a.f[3][0]*b.f[0][1]);
ff.f[1][1]=MIN(a.f[1][1]*b.f[2][0],a.f[2][0]*b.f[1][1]);
ff.f[2][1]=MAX(a.f[2][1]*b.f[1][0],a.f[1][0]*b.f[2][1]);
ff.f[3][1]=MAX(a.f[3][1]*b.f[0][0],a.f[0][0]*b.f[3][1]);
}

ff.flag=FUZZY_FLAG(ff);
return ff;

}
/*------------------------------------------------------------------------- */
void FUZZY_SCANF(struct fuzzy_number * fz)
{int i,j;struct fuzzy_number ff;

scanf("%f %f %f %f %f %f %f %f",&ff.f[0][1],&ff.f[1][1],&ff.f[2][1],&ff.f[3][1],&ff.f[0][0],&ff.f[1][0],&ff.f[2][0],&ff.f[3][0]);
ff.flag=FUZZY_FLAG(ff);

for (j=0;j<2;j++)
for (i=0;i<4;i++) {fz->f[i][j]=ff.f[i][j];};

fz->flag=ff.flag;
}
/*------------------------------------------------------------------------- */
struct fuzzy_number	FUZZY_SIN(struct fuzzy_number a)
{struct fuzzy_number ff;int i,j;

for (j=0;j<2;j++)
for (i=0;i<4;i++) ff.f[i][j]=sin(a.f[i][j]);
ff=FUZZY_SORT(ff);
return ff;

}
/*------------------------------------------------------------------------- */
struct fuzzy_number	FUZZY_SORT(struct fuzzy_number a)
/* The eight values of a fuzzy number _must_ be in ascendant order */
{int i,j;float ftemp[8],temp;struct fuzzy_number ff;

for (j=0;j<2;j++)
for (i=0;i<4;i++) ftemp[j*4+i]=a.f[i][j];


for (i=0;i<7;i++)
for (j=i+1;j<8;j++)
if (ftemp[i]>ftemp[j]) {temp=ftemp[i];ftemp[i]=ftemp[j];ftemp[j]=temp;}

ff=fuzzy_null;
for (i=0;i<4;i++) if (ftemp[i]<0) ff.f[i][1]=ftemp[i];
for (i=4;i<8;i++) if (ftemp[11-i]>0) ff.f[7-i][0]=ftemp[11-i];

ff.flag=FUZZY_FLAG(ff);

return ff;

}
/*------------------------------------------------------------------------- */
struct fuzzy_number	FUZZY_SUM(struct fuzzy_number a,struct fuzzy_number b)
  /*  (a1 U a0) + (b1 U b0) = (a1+b1) U (a1+b0) U (a0+b1) U (a0+b0)  */
{
int i,j;struct fuzzy_number ff,ab; float temp[4];

/* First estimation (add positive parts and negative parts separately */
/*--------------------------------------------  (a1+b1) U (a0+b0)     */
for (j=0;j<2;j++)
for (i=0;i<4;i++) ff.f[i][j]=a.f[i][j]+b.f[i][j];
ff.flag=FUZZY_FLAG(ff);


if (a.flag==0 || b.flag==0) return ff; /* a or b is null */
if (a.flag!=3 && b.flag!=3) return ff; /* a and b have just one part */

/* Here a or b have a positive part _and_ a negative one 
 We take into (account positive_part_of_a + negative _part_of_b) and vice versa. 
We do some approximations, in order to still have at most a bimodal fuzzy set */

/* --------------------------------------------  (a1+b0) */
for (i=0;i<4;i++) temp[i]=a.f[i][1]+b.f[i][0]; 

for (j=0;j<2;j++) for (i=0;i<4;i++) ab.f[i][j]=0;

if (temp[3]<=0) for (i=0;i<4;i++) ab.f[i][1]=temp[i]; /* just a negative part */
if (temp[0]>=0) for (i=0;i<4;i++) ab.f[i][0]=temp[i]; /* just a positive part */

if (temp[0]*temp[3]<0) 
	{ab.f[0][1]=temp[0];ab.f[3][1]=0;ab.f[0][0]=0;ab.f[3][0]=temp[3];
	if (temp[2]<=0) 
		{ab.f[1][1]=temp[1];
		if ((temp[2]+temp[3])/2>0) 
			{ab.f[2][1]=0;ab.f[1][0]=0;ab.f[2][0]=0;}
		else
			{ab.f[2][1]=temp[2];ab.f[3][0]=0;}
		}
	else
		if (temp[1]<=0)
			{ab.f[1][1]=temp[1];ab.f[2][1]=0;ab.f[3][1]=0;ab.f[1][0]=0;ab.f[2][0]=temp[2];}
		else
			if ((temp[0]+temp[1])/2>0) 
			{ab.f[1][1]=0;ab.f[2][1]=0;ab.f[1][0]=0;ab.f[2][0]=temp[2];}
		else
			{ab.f[2][0]=0;ab.f[1][0]=temp[1];ab.f[2][0]=temp[2];};
	
	};


ff.flag=FUZZY_FLAG(ff);return ff;
}
/*------------------------------------------------------------------------- */
float MIN(float a,float b)
{
if (a<b) return a; return b;
}
/*------------------------------------------------------------------------- */
float MAX(float a,float b)
{
if (a>b) return a; return b;
}


